import Environment

data Expr = ExprVar Id
          | ExprLam Id Expr
          | ExprApp Expr Expr
          | ExprConstNum Int
          | ExprConstBool Bool
          | ExprAdd Expr Expr
          | ExprLet Id Expr Expr
          | ExprIf Expr Expr Expr
  deriving Show

data Thunk = TT Expr (Env Thunk)

data Val = VN Int
         | VB Bool
         | VClosure Id Expr (Env Thunk)

instance Show Val where
  show (VN n)           = "VN " ++ show n
  show (VB b)           = "VB " ++ show b
  show (VClosure _ _ _) = "<clausura>"

addVal :: Val -> Val -> Val
addVal (VN a) (VN b) = VN (a + b)
addVal _ _           = error "Los valores no son numéricos."

isTrueVal :: Val -> Bool
isTrueVal (VB b) = b
isTrueVal _      = False

eval :: Expr -> Env Thunk -> Val
eval (ExprVar x)      env =
  case lookupEnv env x of
    TT expr2 env2 -> eval expr2 env2
eval (ExprLam x e)    env = VClosure x e env
eval (ExprApp e1 e2)  env =
  let v1 = eval e1 env
   in case v1 of
        VClosure x e1' envOrig -> eval e1' (extendEnv envOrig x (TT e2 env))
        _ -> error "Lo que se aplica no es una función."
eval (ExprConstNum n)  env = VN n
eval (ExprConstBool b) env = VB b
eval (ExprAdd e1 e2)   env = eval e1 env `addVal` eval e2 env
eval (ExprLet x e1 e2) env = eval e2 (extendEnv env x (TT e1 env))
eval (ExprIf e1 e2 e3) env =
  if isTrueVal (eval e1 env)
   then eval e2 env
   else eval e3 env

ejemplo :: Expr
ejemplo =
  ExprLet "suma" (ExprLam "x"
                   (ExprLam "y"
                     (ExprAdd (ExprVar "x") (ExprVar "y"))))
    (ExprLet "f" (ExprApp (ExprVar "suma") (ExprConstNum 5))
      (ExprLet "x" (ExprConstNum 0)
        (ExprApp (ExprVar "f") (ExprConstNum 3))))

